#include <TimerOne.h>
#include <PS2KeyAdvanced.h>
#include "MorsePWM.h"
#include "LiquidCrystalProgmem.h"

/*  Copyright (c) 2020 Jaroslaw Weglinski based on Lewis library 
    https://git.defproc.co.uk/DefProc/Lewis
    Code licensed under the MIT license.*/

//additional licensing:
//TimerOne: Creative Commons Attribution 3.0 United States License.
//[https://github.com/PaulStoffregen/TimerOne]
//[http://creativecommons.org/licenses/by/3.0/us/]

//Lewis/MorsePWM: the MIT License 
//[https://github.com/DefProc/lewis]
//[https://git.defproc.co.uk/DefProc/Lewis/-/blob/master/LICENSE.md]

//PS2KeyAdvanced: LGPL v.2.1
//[https://github.com/techpaul/PS2KeyAdvanced]
//[https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html]

//LiquidCrystal/LiquidCrystalProgmem: LGPL v.2.1
//[https://github.com/arduino-libraries/LiquidCrystal]
//[https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html]

//firmware version for startup banner
#define FW_VER "Morse 0.2"

//LCD panel connection
#define LCD_RS 12
#define LCD_EN 6
#define LCD_D4 A0
#define LCD_D5 A1
#define LCD_D6 A2
#define LCD_D7 A3

//LCD panel dimensions
#define LCD_ROWS 2
#define LCD_COLS 10

//additional wait time for long LCD operations
#define LCD_LONG_OPER_TIME 100

//DIN keyboard connection
#define KB_DATA 2
#define KB_IRQ  3

//Morse code outputs
#define MORSE_DATA_OUT 10
#define MORSE_AUDIO_OUT 11

//custom character map for up arrow - not present in A00 ROM
const PROGMEM byte upArrow[8] = {
  B00100,
  B01110,
  B11111,
  B00000,
  B00100,
  B01110,
  B11111,
  B00000,
};

//custom character map for down arrow - not present in A00 ROM
const PROGMEM byte downArrow[8] = {
  B00000,
  B11111,
  B01110,
  B00100,
  B00000,
  B11111,
  B01110,
  B00100,
};

MorsePWM Morse;
PS2KeyAdvanced keyboard;
LiquidCrystalProgmem lcd(LCD_RS, LCD_EN, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

void processMorse() {
  Morse.timerISR();
}

//flag if startup banner is currently displayed
bool initial = true;

//lcd cursor position tracking: column, row
uint8_t lcd_c = 0;
uint8_t lcd_r = 0;

void setup() {
  lcd.begin(LCD_COLS, LCD_ROWS);
  delay(LCD_LONG_OPER_TIME);
  lcd.createChar(0, (__FlashStringHelper*)upArrow);
  lcd.createChar(1, (__FlashStringHelper*)downArrow);
  lcd.clear();
  delay(LCD_LONG_OPER_TIME);
  lcd.cursor();
  lcd.setCursor(0, 0);
  lcd.print(F(FW_VER));

  Morse.begin(MORSE_DATA_OUT, MORSE_DATA_OUT, 20, true, MORSE_AUDIO_OUT, 265);

  lcd.setCursor(0, 1);
  delay(LCD_LONG_OPER_TIME);
  lcd.print(Morse.getWPM());
  lcd.print(F("WPM"));
  lcd.blink();

  Timer1.initialize(10000);
  Timer1.attachInterrupt(processMorse);

  keyboard.begin(KB_DATA, KB_IRQ);
  keyboard.setNoBreak(1);
  keyboard.setNoRepeat(1);
}



void loop() {
  unsigned int keycode = 0;
  unsigned char ch;
  bool ctrl_char;

  if ( keyboard.available() ) {
    if ( (keycode = keyboard.read())) {
      ch = (unsigned char)(keycode & 0xff);
      ctrl_char = false;

      if (initial == true || ch == PS2_KEY_ESC) {
        //clear screen on first keystroke or ESC
        Morse.flush();
        initial = false;
        lcd_c = 0;
        lcd_r = 0;
        lcd.clear();
        delay(LCD_LONG_OPER_TIME);
        lcd.setCursor(lcd_c, lcd_r);
      }
      if ((ch <= 0x61 && ch >= 0x1F) || ch == PS2_KEY_UP_ARROW || ch == PS2_KEY_DN_ARROW) {
        if (lcd_c >= LCD_COLS) {
          lcd_c = 0;
          lcd_r++;
          if (lcd_r >= LCD_ROWS) {
            lcd_r = 0;
            lcd.clear();
            delay(LCD_LONG_OPER_TIME);
          }
          lcd.setCursor(lcd_c, lcd_r);
        }
        if (ch == PS2_KEY_UP_ARROW) {
          //increase by 5WPM, limit rounding error accumulation
          Morse.setWPM((Morse.getWPM() / 5 + 1) * 5);
          ch = 0x0; //write up arrow custom character
          ctrl_char = true; //do not encode
        } else if (ch == PS2_KEY_DN_ARROW) {
          //decrease by 5WPM, limit rounding error accumulation
          Morse.setWPM((Morse.getWPM() / 5 - 1) * 5);
          ch = 0x1; //write down arrow custom character
          ctrl_char = true; //do not encode
        }
        lcd.write(ch);
        if (ctrl_char == false)
          Morse.write(ch);
        lcd_c++;
      }
    }
    delay(100);
  }
}
